@@ -47,6 +47,8 @@ struct _DhWindowPriv {
 
 	GtkWidget      *vbox;
 	GtkWidget      *findbar;
+	
+	gint            zoom_level;
 
 	GtkUIManager   *manager;
 	GtkActionGroup *action_group;
@@ -57,6 +59,29 @@ static guint tab_accel_keys[] = {
 	GDK_6, GDK_7, GDK_8, GDK_9, GDK_0
 };
 
+static const
+struct
+{
+	gchar *name;
+	float level;
+}
+zoom_levels[] =
+{
+	{ N_("50%"), 0.7071067811 },
+	{ N_("75%"), 0.8408964152 },
+	{ N_("100%"), 1.0 },
+	{ N_("125%"), 1.1892071149 },
+	{ N_("150%"), 1.4142135623 },
+	{ N_("175%"), 1.6817928304 },
+	{ N_("200%"), 2.0 },
+	{ N_("300%"), 2.8284271247 },
+	{ N_("400%"), 4.0 }
+};
+
+#define ZOOM_MINIMAL	(zoom_levels[0].level)
+#define ZOOM_MAXIMAL	(zoom_levels[8].level)
+#define ZOOM_DEFAULT	(zoom_levels[2].level)
+
 /* People have reported problems with the default values in GConf so I'm
  * adding this to make sure that the window isn't started 1x1 pixels or the
  * paned having size 0
@@ -93,6 +118,12 @@ static void       window_activate_show_search     (GtkAction       *action,
 						   DhWindow        *window);
 static void       window_activate_about           (GtkAction       *action,
 						   DhWindow        *window);
+static void	  window_activate_zoom_default	  (GtkAction 	   *action,
+						   DhWindow	   *window);
+static void	  window_activate_zoom_in	  (GtkAction 	   *action,
+						   DhWindow	   *window);
+static void	  window_activate_zoom_out	  (GtkAction 	   *action,
+						   DhWindow	   *window);
 static void       window_save_state               (DhWindow        *window);
 static void       window_restore_state            (DhWindow        *window);
 static gboolean   window_delete_cb                (GtkWidget       *widget,
@@ -155,6 +186,7 @@ static GtkWindowClass *parent_class = NULL;
 static const GtkActionEntry actions[] = {
 	{ "FileMenu", NULL, N_("_File") },
 	{ "EditMenu", NULL, N_("_Edit") },
+	{ "ViewMenu", NULL, N_("_View") },
 	{ "GoMenu",   NULL, N_("_Go") },
 	{ "HelpMenu", NULL, N_("_Help") },
 
@@ -181,9 +213,11 @@ static const GtkActionEntry actions[] = {
 	  G_CALLBACK (window_activate_preferences) },
 
 	/* Go menu */
-	{ "Back", GTK_STOCK_GO_BACK, NULL, "<alt>Left", NULL,
+	{ "Back", GTK_STOCK_GO_BACK, NULL, "<alt>Left",
+	  N_("Go to the previous page"),
 	  G_CALLBACK (window_activate_back) },
-	{ "Forward", GTK_STOCK_GO_FORWARD, NULL, "<alt>Right", NULL,
+	{ "Forward", GTK_STOCK_GO_FORWARD, NULL, "<alt>Right",
+	  N_("Go to the next page"),
 	  G_CALLBACK (window_activate_forward) },
 
 	{ "ShowContentsTab", NULL, N_("_Contents Tab"), "<ctrl>B", NULL,
@@ -191,6 +225,17 @@ static const GtkActionEntry actions[] = {
 
 	{ "ShowSearchTab", NULL, N_("_Search Tab"), "<ctrl>S", NULL,
 	  G_CALLBACK (window_activate_show_search) },
+	
+	/* View menu */
+	{ "ZoomIn", GTK_STOCK_ZOOM_IN, N_("_Larger Text"), "<ctrl>plus",
+	  N_("Increase the text size"),
+	  G_CALLBACK (window_activate_zoom_in) },
+	{ "ZoomOut", GTK_STOCK_ZOOM_OUT, N_("S_maller Text"), "<ctrl>minus",
+	  N_("Decrease the text size"),
+	  G_CALLBACK (window_activate_zoom_out) },
+	{ "ZoomDefault", GTK_STOCK_ZOOM_100, N_("_Normal size"), "<ctrl>0",
+	  N_("Use the normal text size"),
+	  G_CALLBACK (window_activate_zoom_default) },
 
 	/* About menu */
 	{ "About", GTK_STOCK_ABOUT, NULL, NULL, NULL,
@@ -233,6 +278,15 @@ window_class_init (DhWindowClass *klass)
         object_class = G_OBJECT_CLASS (klass);
 
         object_class->finalize = window_finalize;
+
+	gtk_rc_parse_string ("style \"devhelp-tab-close-button-style\"\n"
+                             "{\n"
+                             "GtkWidget::focus-padding = 0\n"
+                             "GtkWidget::focus-line-width = 0\n"
+                             "xthickness = 0\n"
+                             "ythickness = 0\n"
+                             "}\n"
+                             "widget \"*.devhelp-tab-close-button\" style \"devhelp-tab-close-button-style\"");
 }
 
 static void
@@ -253,6 +307,8 @@ window_init (DhWindow *window)
 
 	accel_group = gtk_ui_manager_get_accel_group (priv->manager);
 	gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);
+	
+	priv->zoom_level = 2;
 
 	priv->main_box = gtk_vbox_new (FALSE, 0);
 	gtk_widget_show (priv->main_box);
@@ -311,7 +367,7 @@ static void
 window_finalize (GObject *object)
 {
 	DhWindow *window = DH_WINDOW (object);
-	DhBase *base = window->priv->base;
+	DhBase   *base = window->priv->base;
 
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 
@@ -604,6 +660,51 @@ window_activate_find (GtkAction *action,
 	gtk_widget_grab_focus (priv->findbar);
 }
 
+static void
+window_activate_zoom_in (GtkAction *action,
+			 DhWindow *window)
+{
+	DhWindowPriv *priv;
+	DhHtml       *html;
+
+	html = window_get_active_html (window);
+	priv = window->priv;
+
+	if (zoom_levels[priv->zoom_level].level < ZOOM_MAXIMAL) {
+		priv->zoom_level++;
+		dh_html_set_zoom (html, zoom_levels[priv->zoom_level].level);
+	}
+}
+
+static void
+window_activate_zoom_out (GtkAction *action,
+			  DhWindow *window)
+{
+	DhHtml       *html;
+	DhWindowPriv *priv;
+
+	html = window_get_active_html (window);
+	priv = window->priv;
+
+	if (zoom_levels[priv->zoom_level].level > ZOOM_MINIMAL) {
+		priv->zoom_level--;
+		dh_html_set_zoom (html, zoom_levels[priv->zoom_level].level);
+	}
+}
+
+static void
+window_activate_zoom_default (GtkAction *action,
+			      DhWindow *window)
+{
+	DhHtml       *html;
+	DhWindowPriv *priv;
+
+	html = window_get_active_html (window);
+	priv = window->priv;
+
+	dh_html_set_zoom (html, ZOOM_DEFAULT);
+}
+
 static void
 window_activate_preferences (GtkAction *action,
 			     DhWindow  *window)
@@ -1083,7 +1184,7 @@ window_open_new_tab (DhWindow    *window,
 	g_object_set_data (G_OBJECT (frame), "html", html);
 
 	label = window_new_tab_label (window, _("Empty Page"));
-	gtk_widget_show (label);
+	gtk_widget_show_all (label);
 
 	g_signal_connect (html, "title-changed",
 			  G_CALLBACK (window_html_title_changed_cb),
@@ -1121,18 +1222,83 @@ window_open_new_tab (DhWindow    *window,
 
 }
 
+static void
+close_button_clicked_cb (GtkButton *button,
+			 DhWindow *window)
+{
+	GtkAction *action_close;
+	
+	action_close = gtk_action_group_get_action (window->priv->action_group,
+						    "Close");
+	window_activate_close (action_close, window);
+}
+
+static void
+tab_label_style_set_cb (GtkWidget *hbox,
+			GtkStyle *previous_style,
+		        gpointer user_data)
+{
+	PangoFontMetrics *metrics;
+	PangoContext     *context;
+	GtkWidget        *button;
+	gint              char_width;
+	gint              h, w;
+
+	context = gtk_widget_get_pango_context (hbox);
+	metrics = pango_context_get_metrics (context,
+					     hbox->style->font_desc,
+					     pango_context_get_language (context));
+
+	char_width = pango_font_metrics_get_approximate_digit_width (metrics);
+	pango_font_metrics_unref (metrics);
+
+	gtk_icon_size_lookup_for_settings (gtk_widget_get_settings (hbox),
+					   GTK_ICON_SIZE_MENU, &w, &h);
+
+	gtk_widget_set_size_request (hbox, 15 * PANGO_PIXELS (char_width) + 2 * w, -1);
+
+	button = g_object_get_data (G_OBJECT (hbox), "close-button");
+	gtk_widget_set_size_request (button, w + 2, h + 2);
+}
+
 static GtkWidget*
 window_new_tab_label (DhWindow *window, const gchar *str)
 {
-	GtkWidget *label;
+	GtkWidget  *hbox, *label, *close_button, *image;
+	
+	hbox = gtk_hbox_new (FALSE, 4);
 
 	label = gtk_label_new (str);
 	gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
 	gtk_label_set_single_line_mode (GTK_LABEL (label), TRUE);
 	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
 	gtk_misc_set_padding (GTK_MISC (label), 0, 0);
+	gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
+	
+	close_button = gtk_button_new ();
+	gtk_button_set_relief (GTK_BUTTON (close_button),
+			       GTK_RELIEF_NONE);
+	gtk_button_set_focus_on_click (GTK_BUTTON (close_button), FALSE);
+	
+	gtk_widget_set_name (close_button, "devhelp-tab-close-button");
+	
+	image = gtk_image_new_from_stock (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
+	/*gtk_widget_set_tooltip_text (close_button, _("Close tab"));*/
+	g_signal_connect (close_button, "clicked",
+			  G_CALLBACK (close_button_clicked_cb), window);
+	
+	gtk_container_add (GTK_CONTAINER (close_button), image);
 
-	return label;
+	gtk_box_pack_start (GTK_BOX (hbox), close_button, FALSE, FALSE, 0);
+	
+	/* Set minimal size */
+	g_signal_connect (hbox, "style-set",
+			  G_CALLBACK (tab_label_style_set_cb), NULL);
+	
+	g_object_set_data (G_OBJECT (hbox), "label", label);
+	g_object_set_data (G_OBJECT (hbox), "close-button", close_button);
+
+	return hbox;
 }
 
 static DhHtml *
@@ -1204,6 +1370,7 @@ window_tab_set_title (DhWindow *window, DhHtml *html, const gchar *title)
 	gint          num_pages, i;
 	GtkWidget    *view;
 	GtkWidget    *page;
+	GtkWidget    *hbox;
 	GtkWidget    *label;
 
 	priv = window->priv;
@@ -1221,10 +1388,11 @@ window_tab_set_title (DhWindow *window, DhHtml *html, const gchar *title)
 
 		/* The html widget is inside a frame. */
 		if (gtk_bin_get_child (GTK_BIN (page)) == view) {
-			label = gtk_notebook_get_tab_label (
+			hbox = gtk_notebook_get_tab_label (
 				GTK_NOTEBOOK (priv->html_notebook), page);
 
-			if (label) {
+			if (hbox) {
+				label = g_object_get_data (G_OBJECT (hbox), "label");
 				gtk_label_set_text (GTK_LABEL (label), title);
 			}
 			break;
